import 'package:flutter/material.dart';

/// field validator interface
abstract class ValidatorState {
  /// dirty
  bool dirty;

  /// if invalid == true validator is invalid validator function should return null
  ///
  /// * if _skipDirty == ture then invalid return false
  /// * if _skipDirty == false then invalid = !dirty
  bool get invalid => _skipDirty ? false : !dirty;
  bool _skipDirty = false;

  /// if true auto validate
  bool autovalidate;

  /// retrun field initial value
  get initialValue;

  /// return to FormFieldValidator compatible functions
  FormFieldValidator get validator;

  /// return to FormFieldSetter compatible functions
  FormFieldSetter get onSaved;
}

typedef FieldValidator<T> = String Function(
  FieldValidatorState<T> validator,
  T value,
);

/// FormField autovalidate support dirty
class FieldValidatorState<T> extends ValidatorState {
  /// if true auto check invalid
  ///
  /// if false you should check invalid and set dirty on your FieldValidator function
  final bool autoinvalid;

  /// retrun field initial value
  @override
  T get initialValue => _initialValue;
  final T _initialValue;

  /// return to FormFieldValidator compatible functions
  @override
  FormFieldValidator get validator =>
      _validator == null ? null : _fieldValidator;
  String _fieldValidator(T v) {
    if (autoinvalid) {
      // 驗證器 無效
      if (invalid) {
        // 初始值 改變 設置 dirty 狀態
        if (_initialValue != v) {
          dirty = true;
          return _validator(this, v);
        }
        return null;
      } else if (!dirty) {
        dirty = true;
      }
    }
    return _validator(this, v);
  }

  final FieldValidator<T> _validator;

  /// return to FormFieldSetter compatible functions
  @override
  FormFieldSetter get onSaved => _onSaved == null ? null : _fieldOnSaved;
  void _fieldOnSaved(T val) => _onSaved(val);
  final FormFieldSetter<T> _onSaved;

  /// Create FieldValidatorState
  FieldValidatorState({
    @required T initialValue,
    FieldValidator<T> validator,
    FormFieldSetter<T> onSaved,
    bool dirty,
    bool autovalidate,
    bool autoinvalid = true,
  })  : this._initialValue = initialValue,
        this._validator = validator,
        this._onSaved = onSaved,
        this.autoinvalid = autoinvalid,
        assert(initialValue != null) {
    if (dirty == null || !dirty) {
      this.dirty = false;
    } else {
      this.dirty = true;
    }

    if (autovalidate == null || !autovalidate) {
      this.autovalidate = false;
    } else {
      this.autovalidate = true;
    }
  }
}

/// Syntactic sugar Map<Field,Validator>
class FormValidatorState {
  Map<String, ValidatorState> fileds;
  FormValidatorState({Map<String, ValidatorState> fileds})
      : this.fileds = fileds;

  /// reset all Field.dirty = false
  void reset() => fileds?.forEach((_, filed) => filed.dirty = false);

  /// if Field exists return filed.autovalidate else or false
  bool autovalidate(String key) {
    if (fileds == null) {
      return false;
    }
    var filed = fileds[key];
    return filed == null ? false : filed.autovalidate;
  }

  /// if Field exists return filed.initialValue else or null
  initialValue(String key) {
    if (fileds == null) {
      return false;
    }
    var filed = fileds[key];
    return filed == null ? null : filed.initialValue;
  }

  /// if Field exists return filed.validator else or null
  FormFieldValidator validator(String key) {
    if (fileds == null) {
      return null;
    }
    var filed = fileds[key];
    return filed == null ? null : filed.validator;
  }

  /// if Field exists return filed.onSaved else or null
  FormFieldSetter onSaved(String key) {
    if (fileds == null) {
      return null;
    }
    var filed = fileds[key];
    return filed == null ? null : filed.onSaved;
  }

  /// call validator for all Field
  bool validate(FormState form) {
    if (fileds == null) {
      return form.validate();
    }
    fileds.forEach((_, filed) => filed._skipDirty = true);
    bool yes = form.validate();
    fileds.forEach((_, filed) => filed._skipDirty = false);
    return yes;
  }
}
